@@ -140,6 +140,16 @@ g_vfs_backend_dav_init (GVfsBackendDav *backend)
 /* ************************************************************************* */
 /* Small utility functions */
 
+static gboolean
+string_to_uint64 (const char *str, guint64 *value)
+{
+  char *endptr;
+
+  *value = g_ascii_strtoull (str, &endptr, 10);
+
+  return endptr != str;
+}
+
 static inline gboolean
 sm_has_header (SoupMessage *msg, const char *header)
 {
@@ -261,6 +271,19 @@ dav_uri_match (SoupURI *a, SoupURI *b, gboolean relax)
   return !diff;
 }
 
+static char *
+dav_uri_encode (const char *path_to_encode)
+{
+  char *path;
+  static const char *allowed_reserved_chars = "/";
+
+  path = g_uri_escape_string (path_to_encode,
+                              allowed_reserved_chars,
+                              FALSE);
+
+  return path;
+}
+
 static gboolean
 message_should_apply_redir_ref (SoupMessage *msg)
 {
@@ -275,6 +298,46 @@ message_should_apply_redir_ref (SoupMessage *msg)
   return TRUE;
 }
 
+
+static SoupURI *
+g_vfs_backend_dav_uri_for_path (GVfsBackend *backend,
+                                const char  *path,
+                                gboolean     is_dir)
+{
+  SoupURI *mount_base;
+  SoupURI *uri;
+  char    *fn_encoded;
+  char    *new_path;
+
+  mount_base = http_backend_get_mount_base (backend);
+  uri = soup_uri_copy (mount_base);
+
+  /* "/" means "whatever mount_base is" */
+  if (!strcmp (path, "/"))
+    return uri;
+
+  /* The mount_base path is escaped already so we need to
+     escape the new path as well */
+  fn_encoded = dav_uri_encode (path);
+
+  /* Otherwise, we append filename to mount_base (which is assumed to
+   * be a directory in this case).
+   *
+   * Add a "/" in cases where it is likely that the url is going
+   * to be a directory to avoid redirections
+   */
+  if (is_dir == FALSE || g_str_has_suffix (path, "/"))
+    new_path = g_build_path ("/", uri->path, fn_encoded, NULL);
+  else
+    new_path = g_build_path ("/", uri->path, fn_encoded, "/", NULL);
+
+  g_free (fn_encoded);
+  g_free (uri->path);
+  uri->path = new_path;
+
+  return uri;
+}
+
 /* redirection */
 static void
 redirect_handler (SoupMessage *msg, gpointer user_data)
@@ -390,6 +453,50 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
     soup_uri_free (new_uri);
 }
 
+static void
+g_vfs_backend_dav_setup_display_name (GVfsBackend *backend)
+{
+  GVfsBackendDav *dav_backend;
+  SoupURI        *mount_base;
+  char           *display_name;
+  char            port[7] = {0, };
+
+  dav_backend = G_VFS_BACKEND_DAV (backend);
+
+#ifdef HAVE_AVAHI
+  if (dav_backend->resolver != NULL)
+    {
+      const char *name;
+      name = g_vfs_dns_sd_resolver_get_service_name (dav_backend->resolver);
+      g_vfs_backend_set_display_name (backend, name);
+      return;
+    }
+#endif
+
+  mount_base = http_backend_get_mount_base (backend);
+
+  if (! soup_uri_uses_default_port (mount_base))
+    g_snprintf (port, sizeof (port), ":%u", mount_base->port);
+
+  if (mount_base->user != NULL)
+    /* Translators: This is the name of the WebDAV share constructed as
+       "WebDAV as <username> on <hostname>:<port>"; the ":<port>" part is
+       the second %s and only shown if it is not the default http(s) port. */
+    display_name = g_strdup_printf (_("WebDAV as %s on %s%s"),
+				    mount_base->user,
+				    mount_base->host,
+				    port);
+  else
+    /* Translators: This is the name of the WebDAV share constructed as
+       "WebDAV on <hostname>:<port>"; The ":port" part is again the second
+       %s and it is only shown if it is not the default http(s) port. */
+    display_name = g_strdup_printf (_("WebDAV on %s%s"),
+				    mount_base->host,
+				    port);
+
+  g_vfs_backend_set_display_name (backend, display_name);
+}
+
 static guint
 g_vfs_backend_dav_send_message (GVfsBackend *backend, SoupMessage *message)
 {
@@ -612,15 +719,17 @@ struct _Multistatus {
   xmlNodePtr root;
 
   SoupURI *target;
+  char    *path;
 
 };
 
 struct _MsResponse {
 
   Multistatus *multistatus;
+  char        *path;
+  gboolean     is_target;
+  xmlNodePtr   first_propstat;
 
-  xmlNodePtr  href;
-  xmlNodePtr  first_propstat;
 };
 
 struct _MsPropstat {
@@ -636,17 +745,21 @@ struct _MsPropstat {
 static gboolean
 multistatus_parse (SoupMessage *msg, Multistatus *multistatus, GError **error)
 {
-  xmlDocPtr  doc;
-  xmlNodePtr root;
+  xmlDocPtr   doc;
+  xmlNodePtr  root;
+  SoupURI    *uri;
 
   doc = parse_xml (msg, &root, "multistatus", error);
 
   if (doc == NULL)
     return FALSE;
 
+  uri = soup_message_get_uri (msg);
+
   multistatus->doc = doc;
   multistatus->root = root;
-  multistatus->target = soup_message_get_uri (msg);
+  multistatus->target = uri;
+  multistatus->path = g_uri_unescape_string (uri->path, "/");
 
   return TRUE;
 }
@@ -655,6 +768,7 @@ static void
 multistatus_free (Multistatus *multistatus)
 {
   xmlFreeDoc (multistatus->doc);
+  g_free (multistatus->path);
 }
 
 static void
@@ -675,6 +789,9 @@ multistatus_get_response (xmlNodeIter *resp_iter, MsResponse *response)
   xmlNodePtr   iter;
   xmlNodePtr   href;
   xmlNodePtr   propstat;
+  SoupURI     *uri;
+  const char  *text;
+  char        *path;
 
   multistatus = xml_node_iter_get_user_data (resp_iter);
   resp_node = xml_node_iter_get_current (resp_iter);
@@ -708,48 +825,37 @@ multistatus_get_response (xmlNodeIter *resp_iter, MsResponse *response)
   if (href == NULL)
     return FALSE;
 
-  response->href = href;
+  text = node_get_content (href);
+
+  if (text == NULL)
+    return FALSE;
+
+  uri = soup_uri_new_with_base (multistatus->target, text);
+
+  if (uri == NULL)
+    return FALSE;
+
+  path = g_uri_unescape_string (uri->path, "/");
+  soup_uri_free (uri);
+
+  response->path = path;
+  response->is_target = path_equal (path, multistatus->path, TRUE);
   response->multistatus = multistatus;
   response->first_propstat = propstat;
 
   return resp_node != NULL;
 }
 
+static void
+ms_response_clear (MsResponse *response)
+{
+  g_free (response->path);
+}
+
 static char *
 ms_response_get_basename (MsResponse *response)
 {
-  const char *text;
-  text = node_get_content (response->href);
-
-  return http_uri_get_basename (text);
-
-}
-
-static gboolean
-ms_response_is_target (MsResponse *response)
-{
-  const char *text;
-  SoupURI    *target;
-  SoupURI    *uri;
-  gboolean    res;
-
-  uri    = NULL;
-  target = response->multistatus->target;
-  text   = node_get_content (response->href);
-
-  if (text == NULL)
-    return FALSE;
-
-  uri = soup_uri_new_with_base (target, text);
-
-  if (uri == NULL)
-    return FALSE;
-
-  res = dav_uri_match (uri, target, TRUE);
-
-  soup_uri_free (uri);
-  
-  return res;
+  return http_path_get_basename (response->path);
 }
 
 static void
@@ -901,7 +1007,7 @@ ms_response_to_file_info (MsResponse *response,
               file_type = parse_resourcetype (node);
               g_file_info_set_file_type (info, file_type);
             }
-          else if (node_has_name (node, "displayname"))
+          else if (node_has_name (node, "displayname") && text)
             {
               g_file_info_set_display_name (info, text);
               have_display_name = TRUE;
@@ -973,6 +1079,59 @@ ms_response_to_file_info (MsResponse *response,
 
 }
 
+static void
+ms_response_to_fs_info (MsResponse *response,
+                        GFileInfo  *info)
+{
+  xmlNodeIter iter;
+  MsPropstat  propstat;
+  xmlNodePtr  node;
+  guint       status;
+  const char *text;
+
+  ms_response_get_propstat_iter (response, &iter);
+  while (xml_node_iter_next (&iter))
+    {
+      status = ms_response_get_propstat (&iter, &propstat);
+
+      if (! SOUP_STATUS_IS_SUCCESSFUL (status))
+        continue;
+
+      for (node = propstat.prop_node->children; node; node = node->next)
+        {
+          if (! node_is_element (node))
+            continue;
+
+          text = node_get_content (node);
+          if (text == NULL)
+            continue;
+
+          if (node_has_name (node, "quota-available-bytes"))
+            {
+              guint64 size;
+
+              if (! string_to_uint64 (text, &size))
+                continue;
+
+              g_file_info_set_attribute_uint64 (info,
+                                                G_FILE_ATTRIBUTE_FILESYSTEM_SIZE,
+                                                size);
+            }
+          else if (node_has_name (node, "quota-used-bytes"))
+            {
+              guint64 size;
+
+              if (! string_to_uint64 (text, &size))
+                continue;
+
+              g_file_info_set_attribute_uint64 (info,
+                                                G_FILE_ATTRIBUTE_FILESYSTEM_FREE,
+                                                size);
+            }
+        }
+    }
+}
+
 static GFileType
 ms_response_to_file_type (MsResponse *response)
 {
@@ -1040,7 +1199,7 @@ propfind_request_new (GVfsBackend     *backend,
   const char  *header_depth;
   GString     *body;
 
-  uri = http_backend_uri_for_filename (backend, filename, depth > 0);
+  uri = g_vfs_backend_dav_uri_for_path (backend, filename, depth > 0);
   msg = soup_message_new_from_uri (SOUP_METHOD_PROPFIND, uri);
   soup_uri_free (uri);
 
@@ -1151,13 +1310,15 @@ stat_location_finish (SoupMessage *msg,
       if (! multistatus_get_response (&iter, &response))
         continue;
 
-      if (ms_response_is_target (&response))
+      if (response.is_target)
         {
           file_type = ms_response_to_file_type (&response);
           res = TRUE;
         }
       else
         child_count++;
+
+      ms_response_clear (&response);
     }
 
   if (res)
@@ -1423,12 +1584,14 @@ g_mount_spec_to_dav_uri (GMountSpec *spec)
   const char     *user;
   const char     *port;
   const char     *ssl;
+  const char     *path;
   gint            port_num;
 
   host = g_mount_spec_get (spec, "host");
   user = g_mount_spec_get (spec, "user");
   port = g_mount_spec_get (spec, "port");
   ssl  = g_mount_spec_get (spec, "ssl");
+  path = spec->mount_prefix;
 
   if (host == NULL || *host == 0)
     return NULL;
@@ -1441,12 +1604,13 @@ g_mount_spec_to_dav_uri (GMountSpec *spec)
     soup_uri_set_scheme (uri, SOUP_URI_SCHEME_HTTP);
 
   soup_uri_set_user (uri, user);
+  soup_uri_set_host (uri, host);
 
   if (port && (port_num = atoi (port)))
     soup_uri_set_port (uri, port_num);
 
-  soup_uri_set_host (uri, host);
-  soup_uri_set_path (uri, spec->mount_prefix);
+  g_free (uri->path);
+  uri->path = dav_uri_encode (path);
 
   return uri;
 }
@@ -1457,6 +1621,7 @@ g_mount_spec_from_dav_uri (GVfsBackendDav *dav_backend,
 {
   GMountSpec *spec;
   const char *ssl;
+  char       *local_path;
 
 #ifdef HAVE_AVAHI
   if (dav_backend->resolver != NULL)
@@ -1499,7 +1664,11 @@ g_mount_spec_from_dav_uri (GVfsBackendDav *dav_backend,
       g_free (port);
     }
 
-  g_mount_spec_set_mount_prefix (spec, uri->path);
+  /* There must not be any illegal characters in the
+     URL at this point */
+  local_path = g_uri_unescape_string (uri->path, "/");
+  g_mount_spec_set_mount_prefix (spec, local_path);
+  g_free (local_path);
 
   return spec;
 }
@@ -1582,7 +1751,6 @@ do_mount (GVfsBackend  *backend,
   gboolean        is_webdav;
   gboolean        res;
   char           *last_good_path;
-  char           *display_name;
   const char     *host;
   const char     *type;
 
@@ -1723,14 +1891,7 @@ do_mount (GVfsBackend  *backend,
   g_vfs_backend_set_mount_spec (backend, mount_spec);
   g_vfs_backend_set_icon_name (backend, "folder-remote");
   
-#ifdef HAVE_AVAHI
-  if (dav_backend->resolver != NULL)
-    display_name = g_strdup (g_vfs_dns_sd_resolver_get_service_name (dav_backend->resolver));
-  else
-#endif
-    display_name = g_strdup_printf (_("WebDAV on %s"), mount_base->host);
-  g_vfs_backend_set_display_name (backend, display_name);
-  g_free (display_name);
+  g_vfs_backend_dav_setup_display_name (backend);
   
   /* cleanup */
   g_mount_spec_unref (mount_spec);
@@ -1817,11 +1978,13 @@ do_query_info (GVfsBackend           *backend,
       if (! multistatus_get_response (&iter, &response))
         continue;
 
-      if (ms_response_is_target (&response))
+      if (response.is_target)
         {
           ms_response_to_file_info (&response, job->file_info);
           res = TRUE;
         }
+
+      ms_response_clear (&response);
     }
 
   multistatus_free (&ms);
@@ -1836,6 +1999,92 @@ do_query_info (GVfsBackend           *backend,
 
 }
 
+static PropName fs_info_propnames[] = {
+  {"quota-available-bytes", NULL},
+  {"quota-used-bytes",      NULL},
+  {NULL,                    NULL}
+};
+
+static void
+do_query_fs_info (GVfsBackend           *backend,
+                  GVfsJobQueryFsInfo    *job,
+                  const char            *filename,
+                  GFileInfo             *info,
+                  GFileAttributeMatcher *attribute_matcher)
+{
+  SoupMessage *msg;
+  Multistatus  ms;
+  xmlNodeIter  iter;
+  gboolean     res;
+  GError      *error;
+
+  g_file_info_set_attribute_string (info,
+                                    G_FILE_ATTRIBUTE_FILESYSTEM_TYPE,
+                                    "webdav");
+
+  if (! (g_file_attribute_matcher_matches (attribute_matcher,
+                                           G_FILE_ATTRIBUTE_FILESYSTEM_SIZE) ||
+         g_file_attribute_matcher_matches (attribute_matcher,
+                                           G_FILE_ATTRIBUTE_FILESYSTEM_FREE)))
+    {
+      g_vfs_job_succeeded (G_VFS_JOB (job));
+      return;
+    }
+
+  msg = propfind_request_new (backend, filename, 0, fs_info_propnames);
+
+  if (msg == NULL)
+    {
+      g_vfs_job_failed (G_VFS_JOB (job),
+                        G_IO_ERROR, G_IO_ERROR_FAILED,
+                        _("Could not create request"));
+
+      return;
+    }
+
+  g_vfs_backend_dav_send_message (backend, msg);
+
+  error = NULL;
+  res = multistatus_parse (msg, &ms, &error);
+
+  if (res == FALSE)
+    {
+      g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
+      g_error_free (error);
+      g_object_unref (msg);
+      return;
+    }
+
+  res = FALSE;
+  multistatus_get_response_iter (&ms, &iter);
+
+  while (xml_node_iter_next (&iter))
+    {
+      MsResponse response;
+
+      if (! multistatus_get_response (&iter, &response))
+        continue;
+
+      if (response.is_target)
+        {
+          ms_response_to_fs_info (&response, info);
+          res = TRUE;
+        }
+
+      ms_response_clear (&response);
+    }
+
+  multistatus_free (&ms);
+  g_object_unref (msg);
+
+  if (res)
+    g_vfs_job_succeeded (G_VFS_JOB (job));
+  else
+    g_vfs_job_failed (G_VFS_JOB (job),
+                      G_IO_ERROR, G_IO_ERROR_FAILED,
+                      _("Response invalid"));
+
+}
 
 /* *** enumerate *** */
 static void
@@ -1890,12 +2139,14 @@ do_enumerate (GVfsBackend           *backend,
       if (! multistatus_get_response (&iter, &response))
         continue;
 
-      if (ms_response_is_target (&response))
-        continue;
+      if (response.is_target == FALSE)
+	{
+	  info = g_file_info_new ();
+	  ms_response_to_file_info (&response, info);
+	  g_vfs_job_enumerate_add_info (job, info);
+	}
 
-      info = g_file_info_new ();
-      ms_response_to_file_info (&response, info);
-      g_vfs_job_enumerate_add_info (job, info);
+      ms_response_clear (&response);
     }
 
   multistatus_free (&ms);
@@ -1964,7 +2215,7 @@ try_open_for_read (GVfsBackend        *backend,
   SoupMessage     *msg;
   SoupURI         *uri;
 
-  uri = http_backend_uri_for_filename (backend, filename, FALSE);
+  uri = g_vfs_backend_dav_uri_for_path (backend, filename, FALSE);
   msg = stat_location_begin (uri, FALSE);
   soup_uri_free (uri);
 
@@ -2033,7 +2284,7 @@ try_create (GVfsBackend *backend,
   /* TODO: if SoupOutputStream supported chunked requests, we could
    * use a PUT with "If-None-Match: *" and "Expect: 100-continue"
    */
-  uri = http_backend_uri_for_filename (backend, filename, FALSE);
+  uri = g_vfs_backend_dav_uri_for_path (backend, filename, FALSE);
   msg = soup_message_new_from_uri (SOUP_METHOD_HEAD, uri);
   soup_uri_free (uri);
 
@@ -2112,7 +2363,7 @@ try_replace (GVfsBackend *backend,
 
 
 
-  uri = http_backend_uri_for_filename (backend, filename, FALSE);
+  uri = g_vfs_backend_dav_uri_for_path (backend, filename, FALSE);
 
   if (etag)
     {
@@ -2245,7 +2496,7 @@ do_make_directory (GVfsBackend          *backend,
   SoupURI     *uri;
   guint        status;
 
-  uri = http_backend_uri_for_filename (backend, filename, TRUE);
+  uri = g_vfs_backend_dav_uri_for_path (backend, filename, TRUE);
   msg = soup_message_new_from_uri (SOUP_METHOD_MKCOL, uri);
   soup_uri_free (uri);
 
@@ -2281,7 +2532,7 @@ do_delete (GVfsBackend   *backend,
 
   error = NULL;
 
-  uri = http_backend_uri_for_filename (backend, filename, FALSE);
+  uri = g_vfs_backend_dav_uri_for_path (backend, filename, FALSE);
   res = stat_location (backend, uri, &file_type, &num_children, &error);
 
   if (res == FALSE)
@@ -2328,12 +2579,12 @@ do_set_display_name (GVfsBackend           *backend,
   char        *dirname;
   guint        status;
 
-  source = http_backend_uri_for_filename (backend, filename, FALSE);
+  source = g_vfs_backend_dav_uri_for_path (backend, filename, FALSE);
   msg = soup_message_new_from_uri (SOUP_METHOD_MOVE, source);
 
   dirname = g_path_get_dirname (filename);
   target_path = g_build_filename (dirname, display_name, NULL);
-  target = http_backend_uri_for_filename (backend, target_path, FALSE);
+  target = g_vfs_backend_dav_uri_for_path (backend, target_path, FALSE);
 
   message_add_destination_header (msg, target);
   message_add_overwrite_header (msg, FALSE);
@@ -2379,15 +2630,6 @@ do_set_display_name (GVfsBackend           *backend,
   soup_uri_free (source);
 }
 
-static gboolean
-try_unmount (GVfsBackend    *backend,
-             GVfsJobUnmount *job,
-             GMountUnmountFlags flags,
-             GMountSource *mount_source)
-{
-  _exit (0);
-}
-
 /* ************************************************************************* */
 /*  */
 static void
@@ -2405,6 +2647,7 @@ g_vfs_backend_dav_class_init (GVfsBackendDavClass *klass)
   backend_class->mount             = do_mount;
   backend_class->try_query_info    = NULL;
   backend_class->query_info        = do_query_info;
+  backend_class->query_fs_info     = do_query_fs_info;
   backend_class->enumerate         = do_enumerate;
   backend_class->try_open_for_read = try_open_for_read;
   backend_class->try_create        = try_create;
@@ -2414,5 +2657,4 @@ g_vfs_backend_dav_class_init (GVfsBackendDavClass *klass)
   backend_class->make_directory    = do_make_directory;
   backend_class->delete            = do_delete;
   backend_class->set_display_name  = do_set_display_name;
-  backend_class->try_unmount       = try_unmount;
 }
